@@ -1,6 +1,6 @@
 /*
  *  Phusion Passenger - https://www.phusionpassenger.com/
- *  Copyright (c) 2011-2017 Phusion Holding B.V.
+ *  Copyright (c) 2011-2018 Phusion Holding B.V.
  *
  *  "Passenger", "Phusion Passenger" and "Union Station" are registered
  *  trademarks of Phusion Holding B.V.
@@ -32,7 +32,7 @@
 
 #define CC_BENCHMARK_POINT(client, req, value) \
 	do { \
-		if (OXT_UNLIKELY(mainConfigCache.benchmarkMode == value)) { \
+		if (OXT_UNLIKELY(mainConfig.benchmarkMode == value)) { \
 			writeBenchmarkResponse(&client, &req); \
 			return; \
 		} \
@@ -61,7 +61,7 @@
 #include <cassert>
 #include <cctype>
 
-#include <Logging.h>
+#include <LoggingKit/LoggingKit.h>
 #include <MessageReadersWriters.h>
 #include <Constants.h>
 #include <ConfigKit/ConfigKit.h>
@@ -71,6 +71,7 @@
 #include <MemoryKit/palloc.h>
 #include <DataStructures/LString.h>
 #include <DataStructures/StringKeyTable.h>
+#include <WrapperRegistry/Registry.h>
 #include <StaticString.h>
 #include <Utils.h>
 #include <Utils/StrIntUtils.h>
@@ -78,12 +79,10 @@
 #include <Utils/JsonUtils.h>
 #include <Utils/HttpConstants.h>
 #include <Utils/Timer.h>
-#include <Core/ApplicationPool/ErrorRenderer.h>
 #include <Core/Controller/Config.h>
 #include <Core/Controller/Client.h>
 #include <Core/Controller/AppResponse.h>
 #include <Core/Controller/TurboCaching.h>
-#include <Core/UnionStation/Context.h>
 
 namespace Passenger {
 
@@ -114,8 +113,8 @@ private:
 	// has enough bits.
 	static const unsigned int MAX_SESSION_CHECKOUT_TRY = 10;
 
-	ControllerMainConfigCache mainConfigCache;
-	ControllerRequestConfigCachePtr requestConfigCache;
+	ControllerMainConfig mainConfig;
+	ControllerRequestConfigPtr requestConfig;
 	StringKeyTable< boost::shared_ptr<Options> > poolOptionsCache;
 
 	HashedStaticString PASSENGER_APP_GROUP_NAME;
@@ -125,7 +124,6 @@ private:
 	HashedStaticString PASSENGER_STICKY_SESSIONS;
 	HashedStaticString PASSENGER_STICKY_SESSIONS_COOKIE_NAME;
 	HashedStaticString PASSENGER_REQUEST_OOB_WORK;
-	HashedStaticString UNION_STATION_SUPPORT;
 	HashedStaticString REMOTE_ADDR;
 	HashedStaticString REMOTE_PORT;
 	HashedStaticString REMOTE_USER;
@@ -144,6 +142,7 @@ private:
 	friend class ResponseCache<Request>;
 	struct ev_check checkWatcher;
 	TurboCaching<Request> turboCaching;
+	ConfigKit::Store *singleAppModeConfig;
 
 	#ifdef DEBUG_CC_EVENT_LOOP_BLOCKING
 		struct ev_prepare prepareWatcher;
@@ -151,6 +150,11 @@ private:
 	#endif
 
 
+	/****** Initialization and shutdown ******/
+
+	void preinitialize();
+
+
 	/****** Stage: initialize request ******/
 
 	struct RequestAnalysis;
@@ -158,8 +162,8 @@ private:
 	void initializeFlags(Client *client, Request *req, RequestAnalysis &analysis);
 	bool respondFromTurboCache(Client *client, Request *req);
 	void initializePoolOptions(Client *client, Request *req, RequestAnalysis &analysis);
-	void fillPoolOptionsFromConfigCaches(Options &options,
-		const ControllerRequestConfigCachePtr &requestConfigCache);
+	void fillPoolOptionsFromConfigCaches(Options &options, psg_pool_t *pool,
+		const ControllerRequestConfigPtr &requestConfigCache);
 	static void fillPoolOption(Request *req, StaticString &field,
 		const HashedStaticString &name);
 	static void fillPoolOption(Request *req, int &field,
@@ -176,7 +180,6 @@ private:
 		const HashedStaticString &name);
 	void createNewPoolOptions(Client *client, Request *req,
 		const HashedStaticString &appGroupName);
-	void initializeUnionStation(Client *client, Request *req, RequestAnalysis &analysis);
 	void setStickySessionId(Client *client, Request *req);
 	const LString *getStickySessionCookieName(Request *req);
 
@@ -206,11 +209,11 @@ private:
 	void writeRequestQueueFullExceptionErrorResponse(Client *client,
 		Request *req, const boost::shared_ptr<RequestQueueFullException> &e);
 	void writeSpawnExceptionErrorResponse(Client *client, Request *req,
-		const boost::shared_ptr<SpawnException> &e);
+		const boost::shared_ptr<SpawningKit::SpawnException> &e);
 	void writeOtherExceptionErrorResponse(Client *client, Request *req,
 		const ExceptionPtr &e);
 	void endRequestWithErrorResponse(Client **c, Request **r,
-		const StaticString &message, const SpawnException *e = NULL);
+		const SpawningKit::SpawnException &e);
 	bool friendlyErrorPagesEnabled(Request *req);
 
 
@@ -222,7 +225,7 @@ private:
 	void sendHeaderToApp(Client *client, Request *req);
 	void sendHeaderToAppWithSessionProtocol(Client *client, Request *req);
 	static void sendBodyToAppWhenAppSinkIdle(Channel *_channel, unsigned int size);
-	unsigned int determineHeaderSizeForSessionProtocol(Request *req,
+	unsigned int determineMaxHeaderSizeForSessionProtocol(Request *req,
 		SessionProtocolWorkingState &state, string delta_monotonic);
 	bool constructHeaderForSessionProtocol(Request *req, char * restrict buffer,
 		unsigned int &size, const SessionProtocolWorkingState &state, string delta_monotonic);
@@ -287,7 +290,6 @@ private:
 	void handleAppResponseBodyEnd(Client *client, Request *req);
 	OXT_FORCE_INLINE void keepAliveAppConnection(Client *client, Request *req);
 	void storeAppResponseInTurboCache(Client *client, Request *req);
-	void finalizeUnionStationWithSuccess(Client *client, Request *req);
 
 
 	/***** Hooks ******/
@@ -349,8 +351,8 @@ protected:
 		const MemoryKit::mbuf &buffer, int errcode);
 	virtual void onNextRequestEarlyReadError(Client *client, Request *req, int errcode);
 	virtual bool shouldDisconnectClientOnShutdown(Client *client);
+	virtual bool shouldAutoDechunkBody(Client *client, Request *req);
 	virtual bool supportsUpgrade(Client *client, Request *req);
-	virtual void onConfigChange(const ConfigKit::Store *oldConfig);
 
 
 	/****** Marked virtual so that unit tests can mock these ******/
@@ -360,16 +362,43 @@ protected:
 
 
 public:
+	typedef ControllerConfigChangeRequest ConfigChangeRequest;
+
 	// Dependencies
 	ResourceLocator *resourceLocator;
+	WrapperRegistry::Registry *wrapperRegistry;
 	PoolPtr appPool;
-	UnionStation::ContextPtr unionStationContext;
 
 
 	/****** Initialization and shutdown ******/
 
-	Controller(ServerKit::Context *context, const ControllerSchema &schema,
-		const Json::Value &initialConfig);
+	Controller(ServerKit::Context *context,
+		const ControllerSchema &schema,
+		const Json::Value &initialConfig,
+		const ConfigKit::Translator &translator1 = ConfigKit::DummyTranslator(),
+		const ControllerSingleAppModeSchema *singleAppModeSchema = NULL,
+		const Json::Value *_singleAppModeConfig = NULL,
+		const ConfigKit::Translator &translator2 = ConfigKit::DummyTranslator()
+		)
+		: ParentClass(context, schema, initialConfig, translator1),
+
+		  mainConfig(config),
+		  requestConfig(new ControllerRequestConfig(config)),
+		  poolOptionsCache(4),
+
+		  turboCaching(),
+		  singleAppModeConfig(NULL),
+		  resourceLocator(NULL)
+		  /**************************/
+	{
+		if (mainConfig.singleAppMode) {
+			singleAppModeConfig = new ConfigKit::Store(*singleAppModeSchema,
+				*_singleAppModeConfig, translator2);
+		}
+
+		preinitialize();
+	}
+
 	virtual ~Controller();
 	virtual void initialize();
 
@@ -380,7 +409,15 @@ public:
 	virtual StaticString getServerName() const;
 
 
-	/****** State inspection and configuration ******/
+	/****** Configuration handling ******/
+
+	bool prepareConfigChange(const Json::Value &updates,
+		vector<ConfigKit::Error> &errors, ControllerConfigChangeRequest &req);
+	void commitConfigChange(ControllerConfigChangeRequest &req)
+		BOOST_NOEXCEPT_OR_NOTHROW;
+
+
+	/****** State and configuration ******/
 
 	unsigned int getThreadNumber() const; // Thread-safe
 	virtual Json::Value inspectStateAsJson() const;
