@@ -1,6 +1,6 @@
 /*
  *  Phusion Passenger - https://www.phusionpassenger.com/
- *  Copyright (c) 2010-2017 Phusion Holding B.V.
+ *  Copyright (c) 2010-2018 Phusion Holding B.V.
  *
  *  "Passenger", "Phusion Passenger" and "Union Station" are registered
  *  trademarks of Phusion Holding B.V.
@@ -38,6 +38,7 @@
 
 #include <boost/config.hpp>
 #include <boost/scoped_ptr.hpp>
+#include <boost/foreach.hpp>
 #include <sys/types.h>
 #include <sys/socket.h>
 #include <sys/stat.h>
@@ -80,26 +81,25 @@
 #include <ConfigKit/SubComponentUtils.h>
 #include <ServerKit/Server.h>
 #include <ServerKit/AcceptLoadBalancer.h>
-#include <ConfigKit/VariantMapUtils.h>
+#include <AppTypeDetector/Detector.h>
 #include <MessageReadersWriters.h>
 #include <FileDescriptor.h>
 #include <ResourceLocator.h>
 #include <BackgroundEventLoop.cpp>
 #include <FileTools/FileManip.h>
+#include <FileTools/PathSecurityCheck.h>
 #include <Exceptions.h>
 #include <Utils.h>
 #include <Utils/Timer.h>
-#include <Utils/IOUtils.h>
 #include <Utils/MessageIO.h>
-#include <Utils/VariantMap.h>
 #include <Core/OptionParser.h>
 #include <Core/Controller.h>
 #include <Core/ApiServer.h>
 #include <Core/Config.h>
 #include <Core/ConfigChange.h>
 #include <Core/ApplicationPool/Pool.h>
-#include <Core/UnionStation/Context.h>
 #include <Core/SecurityUpdateChecker.h>
+#include <Core/TelemetryCollector.h>
 #include <Core/AdminPanelConnector.h>
 
 using namespace boost;
@@ -146,9 +146,9 @@ namespace Core {
 
 		ResourceLocator resourceLocator;
 		RandomGeneratorPtr randomGenerator;
-		UnionStation::ContextPtr unionStationContext;
-		SpawningKit::ConfigPtr spawningKitConfig;
-		SpawningKit::FactoryPtr spawningKitFactory;
+		SpawningKit::Context::Schema spawningKitContextSchema;
+		SpawningKit::ContextPtr spawningKitContext;
+		ApplicationPool2::ContextPtr appPoolContext;
 		PoolPtr appPool;
 		Json::Value singleAppModeConfig;
 
@@ -167,6 +167,7 @@ namespace Core {
 		oxt::thread *prestarterThread;
 
 		SecurityUpdateChecker *securityUpdateChecker;
+		TelemetryCollector *telemetryCollector;
 		AdminPanelConnector *adminPanelConnector;
 		oxt::thread *adminPanelConnectorThread;
 
@@ -177,6 +178,7 @@ namespace Core {
 			  shutdownCounter(0),
 			  prestarterThread(NULL),
 			  securityUpdateChecker(NULL),
+			  telemetryCollector(NULL),
 			  adminPanelConnector(NULL),
 			  adminPanelConnectorThread(NULL)
 			  /*******************/
@@ -192,6 +194,7 @@ namespace Core {
 			delete adminPanelConnectorThread;
 			delete adminPanelConnector;
 			delete securityUpdateChecker;
+			delete telemetryCollector;
 
 			/*******************/
 			/*******************/
@@ -213,6 +216,7 @@ namespace Core {
 
 using namespace Passenger::Core;
 
+static WrapperRegistry::Registry *coreWrapperRegistry;
 static Schema *coreSchema;
 static ConfigKit::Store *coreConfig;
 static WorkingObjects *workingObjects;
@@ -240,7 +244,7 @@ initializePrivilegedWorkingObjects() {
 	if (password.isString()) {
 		wo->controllerSecureHeadersPassword = password.asString();
 	} else if (password.isObject()) {
-		wo->controllerSecureHeadersPassword = strip(readAll(password["path"].asString()));
+		wo->controllerSecureHeadersPassword = strip(unsafeReadFile(password["path"].asString()));
 	}
 }
 
@@ -259,9 +263,9 @@ initializeSingleAppMode() {
 
 	if (coreConfig->get("single_app_mode_app_type").isNull()) {
 		P_DEBUG("Autodetecting application type...");
-		AppTypeDetector detector(NULL, 0);
-		PassengerAppType appTypeEnum = detector.checkAppRoot(appRoot);
-		if (appTypeEnum == PAT_NONE || appTypeEnum == PAT_ERROR) {
+		AppTypeDetector::Detector detector(*coreWrapperRegistry, NULL, 0);
+		AppTypeDetector::Detector::Result result = detector.checkAppRoot(appRoot);
+		if (result.isNull()) {
 			fprintf(stderr, "ERROR: unable to autodetect what kind of application "
 				"lives in %s. Please specify information about the app using "
 				"--app-type and --startup-file, or specify a correct location to "
@@ -271,13 +275,18 @@ initializeSingleAppMode() {
 			exit(1);
 		}
 
-		appType = getAppTypeName(appTypeEnum);
+		appType = result.wrapperRegistryEntry->language;
 	} else {
 		appType = coreConfig->get("single_app_mode_app_type").asString();
 	}
 
 	if (coreConfig->get("single_app_mode_startup_file").isNull()) {
-		startupFile = appRoot + "/" + getAppTypeStartupFile(getAppType(appType));
+		const WrapperRegistry::Entry &entry = coreWrapperRegistry->lookup(appType);
+		if (entry.defaultStartupFiles.empty()) {
+			startupFile = appRoot + "/";
+		} else {
+			startupFile = appRoot + "/" + entry.defaultStartupFiles[0];
+		}
 	} else {
 		startupFile = coreConfig->get("single_app_mode_startup_file").asString();
 	}
@@ -553,39 +562,62 @@ printInfoInThread() {
 }
 
 static void
-dumpDiagnosticsOnCrash(void *userData) {
+dumpOxtBacktracesOnCrash(void *userData) {
+	cerr << oxt::thread::all_backtraces();
+	cerr.flush();
+}
+
+static void
+dumpControllerStatesOnCrash(void *userData) {
 	WorkingObjects *wo = workingObjects;
 	unsigned int i;
 
-	cerr << "### Backtraces\n";
-	cerr << oxt::thread::all_backtraces();
-	cerr.flush();
-
 	for (i = 0; i < wo->threadWorkingObjects.size(); i++) {
 		ThreadWorkingObjects *two = &wo->threadWorkingObjects[i];
-		cerr << "### Request handler state (thread " << (i + 1) << ")\n";
+		cerr << "####### Controller state (thread " << (i + 1) << ") #######\n";
 		cerr << two->controller->inspectStateAsJson();
-		cerr << "\n";
+		cerr << "\n\n";
 		cerr.flush();
 	}
+}
+
+static void
+dumpControllerConfigsOnCrash(void *userData) {
+	WorkingObjects *wo = workingObjects;
+	unsigned int i;
 
 	for (i = 0; i < wo->threadWorkingObjects.size(); i++) {
 		ThreadWorkingObjects *two = &wo->threadWorkingObjects[i];
-		cerr << "### Request handler config (thread " << (i + 1) << ")\n";
+		cerr << "####### Controller config (thread " << (i + 1) << ") #######\n";
 		cerr << two->controller->inspectConfig();
-		cerr << "\n";
+		cerr << "\n\n";
 		cerr.flush();
 	}
+}
 
-	cerr << "### Pool state (simple)\n";
+static void
+dumpPoolStateOnCrash(void *userData) {
+	WorkingObjects *wo = workingObjects;
+
+	cerr << "####### Pool state (simple) #######\n";
 	// Do not lock, the crash may occur within the pool.
 	Pool::InspectOptions options(Pool::InspectOptions::makeAuthorized());
 	options.verbose = true;
 	cerr << wo->appPool->inspect(options, false);
-	cerr << "\n";
+	cerr << "\n\n";
 	cerr.flush();
 
-	cerr << "### mbuf stats\n\n";
+	cerr << "####### Pool state (XML) #######\n";
+	Pool::ToXmlOptions options2(Pool::ToXmlOptions::makeAuthorized());
+	options2.secrets = true;
+	cerr << wo->appPool->toXml(options2, false);
+	cerr << "\n\n";
+	cerr.flush();
+}
+
+static void
+dumpMbufStatsOnCrash(void *userData) {
+	WorkingObjects *wo = workingObjects;
 	cerr << "nfree_mbuf_blockq  : " <<
 		wo->threadWorkingObjects[0].serverKitContext->mbuf_pool.nfree_mbuf_blockq << "\n";
 	cerr << "nactive_mbuf_blockq: " <<
@@ -594,13 +626,6 @@ dumpDiagnosticsOnCrash(void *userData) {
 		wo->threadWorkingObjects[0].serverKitContext->mbuf_pool.mbuf_block_chunk_size << "\n";
 	cerr << "\n";
 	cerr.flush();
-
-	cerr << "### Pool state (XML)\n";
-	Pool::ToXmlOptions options2(Pool::ToXmlOptions::makeAuthorized());
-	options2.secrets = true;
-	cerr << wo->appPool->toXml(options2, false);
-	cerr << "\n\n";
-	cerr.flush();
 }
 
 static void
@@ -621,12 +646,6 @@ onTerminationSignal(EV_P_ struct ev_signal *watcher, int revents) {
 	}
 }
 
-static void
-spawningKitErrorHandler(const SpawningKit::ConfigPtr &config, SpawnException &e, const Options &options) {
-	ApplicationPool2::processAndLogNewSpawnException(e, options, config);
-}
-
-
 static void
 initializeCurl() {
 	TRACE_POINT();
@@ -658,30 +677,26 @@ initializeNonPrivilegedWorkingObjects() {
 	}
 
 	UPDATE_TRACE_POINT();
-	if (!coreConfig->get("ust_router_address").isNull()) {
-		wo->unionStationContext = boost::make_shared<UnionStation::Context>(
-			coreConfig->get("ust_router_address").asString(),
-			"logging",
-			coreConfig->get("ust_router_password").asString());
+	wo->spawningKitContext = boost::make_shared<SpawningKit::Context>(
+		wo->spawningKitContextSchema);
+	wo->spawningKitContext->resourceLocator = &wo->resourceLocator;
+	wo->spawningKitContext->wrapperRegistry = coreWrapperRegistry;
+	wo->spawningKitContext->randomGenerator = wo->randomGenerator;
+	wo->spawningKitContext->integrationMode = coreConfig->get("integration_mode").asString();
+	wo->spawningKitContext->instanceDir = coreConfig->get("instance_dir").asString();
+	if (!wo->spawningKitContext->instanceDir.empty()) {
+		wo->spawningKitContext->instanceDir = absolutizePath(
+			wo->spawningKitContext->instanceDir);
 	}
+	wo->spawningKitContext->finalize();
 
 	UPDATE_TRACE_POINT();
-	wo->spawningKitConfig = boost::make_shared<SpawningKit::Config>();
-	wo->spawningKitConfig->resourceLocator = &wo->resourceLocator;
-	wo->spawningKitConfig->agentConfig = coreConfig->inspectEffectiveValues();
-	wo->spawningKitConfig->errorHandler = spawningKitErrorHandler;
-	wo->spawningKitConfig->unionStationContext = wo->unionStationContext;
-	wo->spawningKitConfig->randomGenerator = wo->randomGenerator;
-	wo->spawningKitConfig->instanceDir = coreConfig->get("instance_dir").asString();
-	if (!wo->spawningKitConfig->instanceDir.empty()) {
-		wo->spawningKitConfig->instanceDir = absolutizePath(
-			wo->spawningKitConfig->instanceDir);
-	}
-	wo->spawningKitConfig->finalize();
-
-	UPDATE_TRACE_POINT();
-	wo->spawningKitFactory = boost::make_shared<SpawningKit::Factory>(wo->spawningKitConfig);
-	wo->appPool = boost::make_shared<Pool>(wo->spawningKitFactory, coreConfig->inspectEffectiveValues());
+	wo->appPoolContext = boost::make_shared<ApplicationPool2::Context>();
+	wo->appPoolContext->spawningKitFactory = boost::make_shared<SpawningKit::Factory>(
+		wo->spawningKitContext.get());
+	wo->appPoolContext->agentConfig = coreConfig->inspectEffectiveValues();
+	wo->appPoolContext->finalize();
+	wo->appPool = boost::make_shared<Pool>(wo->appPoolContext.get());
 	wo->appPool->initialize();
 	wo->appPool->setMax(coreConfig->get("max_pool_size").asInt());
 	wo->appPool->setMaxIdleTime(coreConfig->get("pool_idle_time").asInt() * 1000000ULL);
@@ -726,8 +741,8 @@ initializeNonPrivilegedWorkingObjects() {
 			&wo->singleAppModeConfig,
 			coreSchema->controllerSingleAppMode.translator);
 		two.controller->resourceLocator = &wo->resourceLocator;
+		two.controller->wrapperRegistry = coreWrapperRegistry;
 		two.controller->appPool = wo->appPool;
-		two.controller->unionStationContext = wo->unionStationContext;
 		two.controller->shutdownFinishCallback = controllerShutdownFinished;
 		two.controller->initialize();
 		wo->shutdownCounter.fetch_add(1, boost::memory_order_relaxed);
@@ -835,6 +850,27 @@ initializeSecurityUpdateChecker() {
 	checker->start();
 }
 
+static void
+initializeTelemetryCollector() {
+	return; // disable for now
+	TRACE_POINT();
+	WorkingObjects &wo = *workingObjects;
+
+	Json::Value config = coreConfig->inspectEffectiveValues();
+	TelemetryCollector *collector = new TelemetryCollector(
+		coreSchema->telemetryCollector.schema,
+		coreConfig->inspectEffectiveValues(),
+		coreSchema->telemetryCollector.translator);
+	wo.telemetryCollector = collector;
+	for (unsigned int i = 0; i < wo.threadWorkingObjects.size(); i++) {
+		ThreadWorkingObjects *two = &wo.threadWorkingObjects[i];
+		collector->controllers.push_back(two->controller);
+	}
+	collector->initialize();
+	collector->start();
+	wo.shutdownCounter.fetch_add(1, boost::memory_order_relaxed);
+}
+
 static void
 runAdminPanelConnector(AdminPanelConnector *connector) {
 	connector->run();
@@ -863,6 +899,10 @@ initializeAdminPanelConnector() {
 	connector->resourceLocator = &wo.resourceLocator;
 	connector->appPool = wo.appPool;
 	connector->configGetter = inspectConfig;
+	for (unsigned int i = 0; i < wo.threadWorkingObjects.size(); i++) {
+		ThreadWorkingObjects *two = &wo.threadWorkingObjects[i];
+		connector->controllers.push_back(two->controller);
+	}
 	connector->initialize();
 	wo.shutdownCounter.fetch_add(1, boost::memory_order_relaxed);
 	wo.adminPanelConnector = connector;
@@ -894,48 +934,6 @@ prestartWebApps() {
 	);
 }
 
-/**
- * See warnIfPassengerRootVulnerable()
- */
-static void
-warnIfPathVulnerable(const char *path, string &warnings) {
-	struct stat pathStat;
-
-	if (stat(path, &pathStat) == -1) {
-		P_DEBUG("Vulnerability check skipped: stat error on " << path << " (errno: " << errno << ")");
-		return; // fatal: we need that stat for both checks below
-	}
-
-	// Non-root ownership
-	struct passwd pathOwner;
-	struct passwd *pwdResult;
-
-	boost::shared_array<char> strings;
-	long stringsBufSize = std::max<long>(1024 * 128, sysconf(_SC_GETPW_R_SIZE_MAX));
-	strings.reset(new char[stringsBufSize]);
-	errno = 0;
-	if (getpwuid_r(pathStat.st_uid, &pathOwner, strings.get(), stringsBufSize, &pwdResult) == -1) {
-		P_DEBUG("Vulnerability check (owner) skipped: getpwuid_r error on " << path << " (owner UID: " <<
-				pathStat.st_uid << ", errno: " << errno << ")");
-	} else if (pwdResult == NULL) {
-		P_DEBUG("Vulnerability check (owner) skipped: getpwuid_r empty on " << path << " (owner UID: " <<
-				pathStat.st_uid << ", errno: " << errno << ")");
-	} else if (pathOwner.pw_uid != 0) {
-		warnings.append("\nThe path \"");
-		warnings.append(path);
-		warnings.append("\" can be modified by user \"");
-		warnings.append(pathOwner.pw_name);
-		warnings.append("\" (or applications running as that user). Change the owner of the path to root, or avoid running Passenger as root.");
-	}
-
-	// World writeable access rights
-	if ((pathStat.st_mode & S_IWOTH) != 0) {
-		warnings.append("\nThe path \"");
-		warnings.append(path);
-		warnings.append("\" is writeable by any user (or application). Limit write access on the path to only the root user/group.");
-	}
-}
-
 /*
  * Emit a warning (log) if the Passenger root dir (and/or its parents) can be modified by non-root users
  * while Passenger was run as root (because non-root users can then tamper with something running as root).
@@ -955,18 +953,28 @@ warnIfPassengerRootVulnerable() {
 	}
 
 	string root = workingObjects->resourceLocator.getInstallSpec();
-	string checkPath = absolutizePath(root);
-	// Check the Passenger root and all dirs above it for ownership and world-writeability
-	string warnings;
-	while (!checkPath.empty() && checkPath != "/") {
-		warnIfPathVulnerable(checkPath.c_str(), warnings);
-
-		checkPath = extractDirName(checkPath);
-	}
-	if (!warnings.empty()) {
-		P_WARN("WARNING: potential privilege escalation vulnerability. "
-			PROGRAM_NAME " is running as root, and part(s) of the passenger root path ("
-			<< root << ") can be changed by non-root user(s):" << warnings);
+	vector<string> errors, checkErrors;
+	if (isPathProbablySecureForRootUse(root, errors, checkErrors)) {
+		if (!checkErrors.empty()) {
+			string message = "WARNING: unable to perform privilege escalation vulnerability detection:\n";
+			foreach (string line, checkErrors) {
+				message.append("\n - " + line);
+			}
+			P_WARN(message);
+		}
+	} else {
+		string message = "WARNING: potential privilege escalation vulnerability detected. " \
+			PROGRAM_NAME " is running as root, and part(s) of the " SHORT_PROGRAM_NAME
+			" root path (" + root + ") can be changed by non-root user(s):\n";
+		foreach (string line, errors) {
+			message.append("\n - " + line);
+		}
+		foreach (string line, checkErrors) {
+			message.append("\n - " + line);
+		}
+		message.append("\n\nPlease either fix up the permissions for the insecure paths, or install "
+			SHORT_PROGRAM_NAME " in a different location that can only be modified by root.");
+		P_WARN(message);
 	}
 }
 
@@ -1010,6 +1018,50 @@ reportInitializationInfo() {
 	}
 }
 
+static void
+initializeAbortHandlerCustomerDiagnostics() {
+	if (!Agent::Fundamentals::abortHandlerInstalled()) {
+		return;
+	}
+
+	Agent::Fundamentals::AbortHandlerConfig::DiagnosticsDumper *diagnosticsDumpers
+		= &Agent::Fundamentals::context->abortHandlerConfig.diagnosticsDumpers[0];
+
+	diagnosticsDumpers[0].name = "OXT backtraces";
+	diagnosticsDumpers[0].logFileName = "backtrace_oxt.log";
+	diagnosticsDumpers[0].func = dumpOxtBacktracesOnCrash;
+
+	diagnosticsDumpers[1].name = "controller states";
+	diagnosticsDumpers[1].logFileName = "controller_states.log";
+	diagnosticsDumpers[1].func = dumpControllerStatesOnCrash;
+
+	diagnosticsDumpers[2].name = "controller configs";
+	diagnosticsDumpers[2].logFileName = "controller_configs.log";
+	diagnosticsDumpers[2].func = dumpControllerConfigsOnCrash;
+
+	diagnosticsDumpers[3].name = "pool state";
+	diagnosticsDumpers[3].logFileName = "pool.log";
+	diagnosticsDumpers[3].func = dumpPoolStateOnCrash;
+
+	diagnosticsDumpers[4].name = "mbuf statistics";
+	diagnosticsDumpers[4].logFileName = "mbufs.log";
+	diagnosticsDumpers[4].func = dumpMbufStatsOnCrash;
+
+	Agent::Fundamentals::abortHandlerConfigChanged();
+}
+
+static void
+uninstallAbortHandlerCustomDiagnostics() {
+	if (!Agent::Fundamentals::abortHandlerInstalled()) {
+		return;
+	}
+
+	for (unsigned int i = 0; i < Agent::Fundamentals::AbortHandlerConfig::MAX_DIAGNOSTICS_DUMPERS; i++) {
+		Agent::Fundamentals::context->abortHandlerConfig.diagnosticsDumpers[i].func = NULL;
+	}
+	Agent::Fundamentals::abortHandlerConfigChanged();
+}
+
 static void
 mainLoop() {
 	TRACE_POINT();
@@ -1020,9 +1072,6 @@ mainLoop() {
 			&& maxCpus <= CPU_SETSIZE;
 	#endif
 
-	Agent::Fundamentals::context->abortHandlerConfig.diagnosticsDumper = dumpDiagnosticsOnCrash;
-	Agent::Fundamentals::abortHandlerConfigChanged();
-
 	for (unsigned int i = 0; i < wo->threadWorkingObjects.size(); i++) {
 		ThreadWorkingObjects *two = &wo->threadWorkingObjects[i];
 		two->bgloop->start("Main event loop: thread " + toString(i + 1), 0);
@@ -1106,6 +1155,20 @@ apiServerShutdownFinished(Core::ApiServer::ApiServer *server) {
 	serverShutdownFinished();
 }
 
+static void
+telemetryCollectorAsyncShutdownThreadMain() {
+	WorkingObjects *wo = workingObjects;
+	wo->telemetryCollector->stop();
+	serverShutdownFinished();
+}
+
+static void
+asyncShutdownTelemetryCollector() {
+	oxt::thread(telemetryCollectorAsyncShutdownThreadMain,
+		"Telemetry collector shutdown",
+		512 * 1024);
+}
+
 /* Wait until the watchdog closes the feedback fd (meaning it
  * was killed) or until we receive an exit message.
  */
@@ -1127,8 +1190,7 @@ waitForExitEvent() {
 	TRACE_POINT();
 	if (syscalls::select(largestFd + 1, &fds, NULL, NULL, NULL) == -1) {
 		int e = errno;
-		Agent::Fundamentals::context->abortHandlerConfig.diagnosticsDumper = NULL;
-		Agent::Fundamentals::abortHandlerConfigChanged();
+		uninstallAbortHandlerCustomDiagnostics();
 		throw SystemException("select() failed", e);
 	}
 
@@ -1164,6 +1226,9 @@ waitForExitEvent() {
 		if (wo->apiWorkingObjects.apiServer != NULL) {
 			wo->apiWorkingObjects.bgloop->safe->runLater(shutdownApiServer);
 		}
+		if (wo->telemetryCollector != NULL) {
+			asyncShutdownTelemetryCollector();
+		}
 		if (wo->adminPanelConnector != NULL) {
 			wo->adminPanelConnector->asyncShutdown();
 		}
@@ -1175,8 +1240,7 @@ waitForExitEvent() {
 			&fds, NULL, NULL, NULL) == -1)
 		{
 			int e = errno;
-			Agent::Fundamentals::context->abortHandlerConfig.diagnosticsDumper = NULL;
-			Agent::Fundamentals::abortHandlerConfigChanged();
+			uninstallAbortHandlerCustomDiagnostics();
 			throw SystemException("select() failed", e);
 		}
 
@@ -1192,8 +1256,7 @@ cleanup() {
 	P_DEBUG("Shutting down " SHORT_PROGRAM_NAME " core...");
 	wo->appPool->destroy();
 
-	Agent::Fundamentals::context->abortHandlerConfig.diagnosticsDumper = dumpDiagnosticsOnCrash;
-	Agent::Fundamentals::abortHandlerConfigChanged();
+	uninstallAbortHandlerCustomDiagnostics();
 
 	for (unsigned i = 0; i < wo->threadWorkingObjects.size(); i++) {
 		ThreadWorkingObjects *two = &wo->threadWorkingObjects[i];
@@ -1202,6 +1265,11 @@ cleanup() {
 	if (wo->apiWorkingObjects.apiServer != NULL) {
 		wo->apiWorkingObjects.bgloop->stop();
 	}
+	if (wo->telemetryCollector != NULL
+	&& !coreConfig->get("telemetry_collector_disabled").asBool())
+	{
+		wo->telemetryCollector->runOneCycle(true);
+	}
 	wo->appPool.reset();
 	for (unsigned i = 0; i < wo->threadWorkingObjects.size(); i++) {
 		ThreadWorkingObjects *two = &wo->threadWorkingObjects[i];
@@ -1252,12 +1320,14 @@ runCore() {
 		initializeCurl();
 		initializeNonPrivilegedWorkingObjects();
 		initializeSecurityUpdateChecker();
+		initializeTelemetryCollector();
 		initializeAdminPanelConnector();
 		prestartWebApps();
 
 		UPDATE_TRACE_POINT();
 		warnIfPassengerRootVulnerable();
 		reportInitializationInfo();
+		initializeAbortHandlerCustomerDiagnostics();
 		mainLoop();
 
 		UPDATE_TRACE_POINT();
@@ -1309,19 +1379,30 @@ parseOptions(int argc, const char *argv[], ConfigKit::Store &config) {
 	}
 }
 
+static void
+loggingKitPreInitFunc(Json::Value &loggingKitInitialConfig) {
+	loggingKitInitialConfig = manipulateLoggingKitConfig(*coreConfig,
+		loggingKitInitialConfig);
+}
+
 int
 coreMain(int argc, char *argv[]) {
 	int ret;
 
-	coreSchema = new Schema();
+	coreWrapperRegistry = new WrapperRegistry::Registry();
+	coreWrapperRegistry->finalize();
+	coreSchema = new Schema(coreWrapperRegistry);
 	coreConfig = new ConfigKit::Store(*coreSchema);
 	initializeAgent(argc, &argv, SHORT_PROGRAM_NAME " core",
 		*coreConfig, coreSchema->loggingKit.translator,
-		parseOptions, NULL, 2);
+		parseOptions, loggingKitPreInitFunc, 2);
 
+#if !BOOST_OS_MACOS
 	restoreOomScore(coreConfig->get("oom_score").asString());
+#endif
 
 	ret = runCore();
 	shutdownAgent(coreSchema, coreConfig);
+	delete coreWrapperRegistry;
 	return ret;
 }
